/*
*	active.c -- Routines to determine number of active fibers.
*  HISTORY
*	(??)		GDS- Created.
*	(4-7-87)	GDS - Added TEST and main().
*	(5-24-89)	TOR - adapted for NeXT.
*/

#import <math.h>
#import "ccf.h"

extern double	HmSqddB(double f, double fm);
extern double	NoiseVarHmdB(double spl, double fl, double fh, double fm);
extern double	threshold(double f);
extern double	NoiseVarHm(double spl, double fl, double fh, double fm);
extern double	HmSqd(double f, double fm);


/******************************************************************
	active() returns the fraction of fibers with CF fm that will be active
	(i.e. firing rate = 200/sec) when the average power after Hm(f) is 
	filteredSPL (in dB SPL).

	Since I'm changing the active fiber region routine to account for IIDs,
	assume that the routines that call active() take IID into account.  Let 
	active() just return the fraction of active fibers
 ******************************************************************/
double active(double filteredSPL, double fm)
{
	if ( (fm = (filteredSPL - threshold(fm))/40.) >= 1.0 ) return 1.0;
	if (fm <= 0.0) return 0.0;
	return fm;		/* fm is now the fraction of active fibers */
}



/********************************************************************
	ToneActive() will return a structure, Fraction, that contains the
	fractions of doubly-, singly-, and spontaneously-active fibers with
	CF = fm, stimulated	by a tone of freq. f and power SPL dB.
 ********************************************************************/
Fraction ToneActive(double SPL, double IID, double f, double fm)
{
	double		fractionL, fractionR;
	Fraction	fract;
	
	fractionL = active(SPL - fabs(IID) + HmSqddB(f, fm), fm);
	fractionR = active(SPL + HmSqddB(f, fm), fm);
	
	fract.n2 = min(fractionL, fractionR);
	if (fractionL != fractionR)
		fract.n1 = max(fractionL, fractionR) - fract.n2;
	else
		fract.n1 = 0;
	fract.n0 = 1 - fract.n2 - fract.n1;
	
	return fract;	
	
	/* The old ToneActive ignored IIDs: */
	/*	return active(SPL + HmSqddB(f, fm), fm); */
}



/********************************************************************
	AMtoneActive() will return a structure, Fraction, that contains the
	fractions of doubly-, singly-, and spontaneously-active fibers with
	CF = fm, stimulated	by a tone of freq. f and power SPL dB.
 ********************************************************************/
Fraction AMtoneActive(double powerL, double powerR, double IID, double fm)
/* power - unattenuated sum of filtered coefficients */
{
	double		fractionL, fractionR;
	Fraction	fract;
	
	fractionL = active(powerL - IID, fm);
	fractionR = active(powerR, fm);
	
	fract.n2 = min(fractionL, fractionR);
	if (fractionL != fractionR)
		fract.n1 = max(fractionL, fractionR) - fract.n2;
	else
		fract.n1 = 0;
	fract.n0 = 1 - fract.n2 - fract.n1;
	
	return fract;	
}



Fraction NoiseActive(double SPL, double IID, double fl, double fh, double fm)
/*	SPL - overall stimulus level (dB SPL)	*/
{
	double		fractionL, fractionR;
	Fraction	fract;
	
	fractionL = active(NoiseVarHmdB(SPL - fabs(IID), fl, fh, fm), fm);
	fractionR = active(NoiseVarHmdB(SPL, fl, fh, fm), fm);
	
	fract.n2 = min(fractionL, fractionR);
	if (fractionL!=fractionR)
		fract.n1 = max(fractionL, fractionR) - fract.n2;
	else
		fract.n1 = 0;
	fract.n0 = 1 - fract.n2 - fract.n1;
	
	return fract;	
	
	/* The old NoiseActive ignored IIDs: */
	/* return active(NoiseVarHmdB(SPL, fl, fh, fm), fm); */
}



double ToneNoiseActive(double IID, double TSPL, double Tf, double NSPL,
					   double Nfl, double Nfh, double fm)
{
	return active(10.*log10(pow(10.0, TSPL/10.)*HmSqd(Tf, fm)
						+ NoiseVarHm(NSPL, Nfl, Nfh, fm)), fm);
}

Fraction RaatNoiseActive(double SPL, double IID, double fm)
/*	SPL - variance/power of _filtered_ Raatgever noise (dB)	*/
{
	double		fractionL, fractionR;
	Fraction	fract;
	
	fractionL = active(SPL - IID, fm);
	fractionR = active(SPL, fm);
	
	fract.n2 = min(fractionL, fractionR);
	if (fractionL!=fractionR)
		fract.n1 = max(fractionL, fractionR) - fract.n2;
	else
		fract.n1 = 0;
	fract.n0 = 1 - fract.n2 - fract.n1;
	
	return fract;	
	
	/* The old code in RaatNoise() ignored IIDs (It's now a routine here)	*/
	/* Frac = active(10.0*log10(Varm), freq);								*/
	
}
