/*
*	ColburnHm.c
*  HISTORY
*	(2-14-87)	GDS - changed arg.s of NoiseCCF to use N0 instead
*			of varm.  (Made corresponding change in noise.c)
*	(3-21-87)	GDS - changed return value from NoiseCCF to return
*			the 2.*halfvar (halfvar = <sum of> mag) from
*			"return	NoiseVarHm(SPL, fl, fh, fm)".
*	(5-19-87)	GDS - Added ALPHA parameter.  Default of 4.0 is same
*			as it was before, so all prev. predictions were valid
*			ALPHA = 4.0 predictions.(No date string change because
*			prev. pred.'s valid).
*	(7-1-87)	GDS - Added "FreqThresh" variable.  Default is same
*			as it was (800 Hz).  Previous predictions are correct
*			so date string not changed.
*	(7-14-88)	TOR - fixed fft bug by /2 in fft and /NREALPTS in 
*			invfft.
*	(8-26-88)	TOR - wrote Hm function.
*	(9-21-88)	TOR - wrote bilsenCCF.
*	(5-24-89)	TOR - adapted for NeXT.
*/

#import <stdio.h>
#import <math.h>
#import "ccf.h"
#import "Nerve.h"
#import "BiDis.h"

extern int errno;

#define WARNING		1

extern double	sinc(double x);			/* sin(PI*x)/(PI*x) */
extern fft(float *b,int p,int q,int d);

#define TWOPI		(2.0*PI)
#define RADPERDEG	(PI/180.0)

#define	FFT		1	/* Direction flags */
#define INVFFT		0	/* for fft() routine. */

static float	filter[NREALPTS];

#define ALPHA		4.0
#define FreqThresh	800.0

/*	alpha() returns alpha slope parameter of Colburn Hm(f) filters.	*/
double alpha(double fm)		/* Fiber's char. freq. (Hz) */
{
	if (fm <= FreqThresh) return ALPHA;
	return ALPHA*fm/FreqThresh;
}	

/*	Hm() returns Hm(f); using Colburn's Hm(f)	*/
double Hm(f, fm)
	double f, fm;	/* input frequency (Hz), fiber's char. freq. (Hz) */
{
	if (f <= fm)  return pow(f/fm, alpha(fm));
	return pow(fm/f, 2.*alpha(fm));
}


/*	HmSqd() returns pow(Hm(f), 2.); using Colburn's Hm(f)	*/
double HmSqd(f, fm)
	double f, fm;	/* input frequency (Hz), fiber's char. freq. (Hz) */
{
	if (f <= fm)  return pow(f/fm, 2.*alpha(fm));
	return pow(fm/f, 4.*alpha(fm));
}

double HmSqddB(double f, double fm)		/* (Hz), (Hz) */
{
	if (f <= fm) return 20.*alpha(fm)*log10(f/fm);
	return 40.*alpha(fm)*log10(fm/f);
}

/*************************************************************************
	NoiseVarHm() returns the variance of the output of Hm(f) when
	the input is noise with a rectangular spectrum such that
	var{input noise} = N0*(fh-fl)
 *************************************************************************/
double NoiseVarHm(SPL, fl, fh, fm)
	double SPL;	/* Level of input noise (dB SPL) */
	double fl, fh;	/* low and high freq.s (Hz) of input noise spectrum */
	double fm;	/* characteristic frequency of fiber (Hz) */
{
	double a = alpha(fm), a4, a2, N0 = pow(10.0, SPL/10.)/(fh-fl);
	a4 = 1.-4.*a;
	a2 = 2.*a+1.;

	if (fm <= fl)
		return N0*fm*(pow(fh/fm, a4)-pow(fl/fm, a4))/a4;
	if (fm >= fh)
		return N0*fm*(pow(fh/fm, a2)-pow(fl/fm,a2))/a2;
	return N0*fm*((1.-pow(fl/fm, a2))/a2 + (pow(fh/fm, a4)-1.)/a4);
} 

double NoiseVarHmdB(SPL, fl, fh, fm)
	double SPL;	/* Level of input noise (dB SPL) */
	double fl, fh;	/* low and high freq.s (Hz) of input noise spectrum */
	double fm;	/* characteristic frequency of fiber (Hz) */
{
	double a = alpha(fm), a4, a2;
	a4 = 1. - 4.*a;
	a2 = 2.*a + 1.;

	if (fm <= fl)
		return	SPL + 10. * ( log10(fm/(fh-fl)) + log10((pow(fh/fl, a4)-1)/a4)
							  - a4*log10(fm/fl) );
	if (fm >= fh)
		return	SPL + 10. * ( log10(fm/(fh-fl)) + log10((pow(fh/fl, a2)-1)/a2)
							  - a2*log10(fm/fl) );
	return SPL + 10. * log10( fm*((1.-pow(fl/fm, a2))/a2
							  + (pow(fh/fm, a4)-1.)/a4)/(fh-fl) );
} 

/****************************************************************************
	NoiseCCF() computes the unrectified crosscorr. func. between the
	output of a delay filter (td, ps) and its input when its input is the
	response of Hm(f) to noise.
	The stimulus noise has a flat spectral density from fl to fh.
	NoiseCCF() returns the variance of the noise output by Hm(f).
	TOR- didn't insert /NREALPTS in invfft because multfft not used.
	should still work.
 ****************************************************************************/
double NoiseCCF(SPL, fl, fh, td, ps, fm, outp, tup, tstep)
	double	SPL,		/* Overall level of input noise (dB SPL) */
	fl,		/* Lower frequency (Hz) bound of input */
	fh,		/* Upper frequency (Hz) bound of input */
	td,		/* Time delay (msec) */
	ps,		/* Phase shift (degrees) */
	fm,		/* Characteristic Freq. (Hz) */
	*outp,		/* Array for results */
	tstep;		/* Time step (msec) */
	int	tup;		/* Valid range for outp[t] is -tup<= t<= tup*/
{
	register int f;
	double freq, phase, mag, halfN0;
	double fstep = 1000./(NREALPTS*tstep);	/* Hz */
	double halfvar = 0.0;

	halfN0 = 0.5*pow(10.0, SPL/10.0)/(fh-fl);	/* Power/half-Hz */
	ps *= -RADPERDEG;	/* Degrees to -Radians */
	td *= 2.*PI/1000.;	/* msec to sec-Radians */

	/* Fill filter fft with freq. response */
	for (f = 0, freq = 0.0; f < NREALPTS;
	     f+= 2, freq += fstep) {
		if (freq <= fl || freq >= fh)
			filter[f] = filter[f+1] = 0.;
		else {
			phase = ps -freq*td;
			mag = halfN0 * HmSqd(freq, fm) * fstep;
			halfvar += mag;
			filter[f] = mag * cos(phase);
			filter[f+1] = mag * sin(phase);
		}
	}
	filter[1] = 0.;
	errno = 0;	/* clear any range errors */

	fft(filter, NCOMPPTS, LOGNCOMPPTS, INVFFT);

	outp[0] = filter[0];
	for(f = 1; f <= tup; f++) {
		outp[f] = filter[f];
		outp[-f] = filter[NREALPTS-f];
	}

	return 2.*halfvar;
}


/* Creates bilsen noise ccf function. mostly copied from noiseCCF	*/
double BilsenCCF(SPL, fl, fh, tau, g, fm, outp, tup, tstep)
	double	SPL,		/* Overall level of input noise (dB SPL) */
		fl,		/* Lower frequency (Hz) bound of input */
		fh,		/* Upper frequency (Hz) bound of input */
		tau,		/* Time delay (msec) */
		g,		/* slope factor */
		fm,		/* Characteristic Freq. (Hz) */
		*outp,		/* Array for results */
		tstep;		/* Time step (msec) */
	int	tup;		/* Valid range for outp[t] is -tup<= t<= tup*/
{
	register int f,i;
	double freq, phase, mag, halfN0,ex1,ex2;
	double fstep = 1000./(NREALPTS*tstep);	/* Hz */
	double halfvar = 0.0;

	halfN0 = 0.5*pow(10.0, SPL/10.0)/(fh-fl);	/* Power/half-Hz */
	tau *= 2.*PI/1000.;	/* msec to sec-radians */
	/* Fill filter fft with freq. response */
	for (f = 0, freq = 0.0; f < NREALPTS;
	     f+= 2, freq += fstep) {
		if (freq <= fl || freq >= fh)
			filter[f] = filter[f+1] = 0.;
		else {
			ex1 = g * sin(freq*tau);
			ex2 = 1.0 - g * cos(freq*tau);
			phase = -freq*tau - 2.*atan2(ex1,ex2);
			mag = halfN0 * HmSqd(freq, fm) * fstep;
			halfvar += mag;
			filter[f] = mag * cos(phase);
			filter[f+1] = mag * sin(phase);
/*			printf("%f %f\n",freq,sqrt(filter[f]*filter[f]+
				filter[f+1]*filter[f+1])); */
		}
/*	printf("%f %f\n",freq,atan2(filter[f+1],filter[f])); */
	}
	filter[1] = 0.;
	errno = 0;	/* clear any range errors */
	fft(filter, NCOMPPTS, LOGNCOMPPTS, INVFFT);
	outp[0] = filter[0];
	for(f = 1; f <= tup; f++) {
		outp[f] = filter[f];
		outp[-f] = filter[NREALPTS-f];
	}
	return 2.*halfvar;
}

