/*
*	rectify.c
*  HISTORY
*	(4-6-87)	GDS - Added RectCoeff[], and RectMaxCoeff.
*	(4-27-87)	GDS - Made KMAX and MMAX constants in RectVars[] and
*			made default KMAX = 10 instead of 6.
*	(6-30-88)	TOR - added AMRectify().
*	(5-24-89)	TOR - adapted for NeXT.
*/

#import <stdio.h>
#import <math.h>
#import "ccf.h"
#import "Nerve.h"

extern int		errno;
extern double	Gamma(double x);

/* m = 0, 1, 2,..., MMAX-1, where m is the index for the tone  terms */
/* k = 0, 1, 2,..., KMAX-1, where k is the index for the noise terms */
#define MMAX	15
#define KMAX	15

#define Power	3.0		/* half-wave rectifier power */

#define Fac22 6227020800.*14.*15.*16.*17.*18.*19.*20.*21.*22.

static int Kmax = KMAX;
static int Mmax = MMAX;

static double fac[] = {1., 1., 2., 6., 24., 120., 720., 5040., 40320.,
362880., 3628800., 39916800., 479001600., 6227020800., 6227020800.*14.,
6227020800.*14.*15., 6227020800.*14.*15.*16., 6227020800.*14.*15.*16.*17.,
6227020800.*14.*15.*16.*17.*18., 6227020800.*14.*15.*16.*17.*18.*19.,
6227020800.*14.*15.*16.*17.*18.*19.*20.,
6227020800.*14.*15.*16.*17.*18.*19.*20.*21.,
Fac22, Fac22*23., Fac22*23.*24., Fac22*23.*24.*25., Fac22*23.*24.*25.*26.,
Fac22*23.*24.*25.*26.*27., Fac22*23.*24.*25.*26.*27.*28.,
Fac22*23.*24.*25.*26.*27.*28.*29., Fac22*23.*24.*25.*26.*27.*28.*29.*30.,
Fac22*23.*24.*25.*26.*27.*28.*29.*30.*31.,
Fac22*23.*24.*25.*26.*27.*28.*29.*30.*31.*32.,
Fac22*23.*24.*25.*26.*27.*28.*29.*30.*31.*32.*33.};


static double Fac(int k)
{
	double p = 1.;
	if (k == 0 || k == 1)
		return 1.;
	while (k > 1)
		p *= k--;
	return p;
}

#define SUMTOL	1.e-5		/* Stopping condition for summations */

/* confluent hypergeometric function (cf. Davenport and Root (1958), p.303) */
static double F11(double a, double c, double z)
{
	double sum, rfac, el;
	int r = 1;
	
	sum = rfac = el = 1.0;
	for (r = 1; fabs(el/sum) > SUMTOL; r++) {
		el *= a * z / (c * r);
		sum += el;
		c += 1.0;
		a += 1.0;
	}
	return sum;
}

static double h(int m, int k, double VarT, double VarN, double A)
{
	double num, denom, tmp, Tpow;
	tmp = (m + k - Power) / 2.;

	if (tmp > 0.0 && (double)((int) tmp) == tmp)
		return 0.0; 

	if (VarT <= 0.0) {
		if (m == 0)
			Tpow = 1.0;
		else
			return 0.0;
	} else
		Tpow = Pow(VarT/VarN, (double) m/2.);

	num = A * Gamma(Power + 1.) * Tpow * F11(tmp, m + 1., -VarT / VarN);
	denom = 2. * fac[m] * Gamma(1. - tmp) *Pow(VarN, - Power/2.)
		  * Pow(.5, (k - Power)/2.);
		  
	return num / denom;
}


double AGC(double VarT, double VarN)
{
	return NERVE_ACTIVE_RATE / h(0, 0, VarT, VarN, 1.0);
}

static double hArr[MMAX][KMAX];
static double PrvVarT = -1., PrvVarN = -1., PrvPow = -1.;

static double CalcCoefficients(double VarT, double VarN)
{
	register int m, k;
	double A;

	if (VarT == PrvVarT && VarN == PrvVarN && Power == PrvPow)
		return 0;

	PrvVarT = VarT; PrvVarN = VarN; PrvPow = Power;

	A = AGC(VarT, VarN);

	for (m = 0; m < MMAX; m++) {
		for (k = 0; k < KMAX; k++) {
			hArr[m][k] = h(m, k, VarT, VarN, A);
		}
		if (VarT <= 0.0) break;
	}
	return 1;
}

double RectFunc(double x, double AG)
{
	if (x <= 0.0)
		return 0.0;
	return AG * pow(x, Power);
}

int nrectify(double VarT, double Tf, double Ttd, double VarN, double *Np,
	 	 int tup, double tstep, double *op)
{
	register int m = 0, k, t;
	int	kmax = Kmax, mmax = Mmax;
	double tau, tmp, cs, arg;

	if (kmax > KMAX)
		kmax = KMAX;
	if (mmax > MMAX)
		mmax = MMAX;
		
	Kmax = kmax;
	Mmax = mmax;

	CalcCoefficients(VarT, VarN);

	/* contribution from DC (h00) */
	for (t = -tup; t <= tup;)
		op[t++] = DOUBLY_ACTIVE_PROD;

	/* contribution from noise with noise (h0k) */
	for (k = 1; k < kmax; k++) {
		tmp = hArr[0][k] * hArr[0][k] / fac[k];
		
		if (tmp <= 0.0)
			continue;
			
		for (t = -tup; t <= tup; t++)
			op[t] += tmp * pow(Np[t]/VarN, (double)k);
	}
	
	if (VarT <= 0.0) return 0;	/* No tone ==> Done */


	Tf *= 2. * PI / 1000.;
	/* contribution from tone with tone (hm0) */
	for (m = 1, arg = Tf; m < mmax; m++, arg += Tf) {
		tmp = 2. *hArr[m][0] * hArr[m][0];
		if (tmp <= 0.0) continue;
		t = -tup;
		for (tau = t * tstep; t <= tup; t++, tau += tstep)
			op[t] += tmp * cos(arg*(tau - Ttd));
	}

	/* contribution from tone with noise (hmk: m,k > 0) */
	t = -tup;
	for (tau = t*tstep; t <= tup; t++, tau += tstep) {
		for (m = 1, arg = Tf; m < mmax; m++, arg += Tf) {
			cs = cos(arg*(tau - Ttd));
			for (k = 1, tmp = Np[t]/VarN; k < kmax; k++, tmp *= Np[t]/VarN) {
				op[t] += tmp *cs *2. *hArr[m][k] *hArr[m][k] /fac[k];
			}
		}
	}
	return 0;
}

/* Normalized Fourier series coeff. of rectified wave (p123, Shear's thesis */
#define MAXN	8
#define MAXP	3

static double ToneCoeff[MAXP][MAXN+1] = {	/* |Cn/C0|^2 */
	{1., PI*PI/16., 1./9., 0., 1./225., 0., 1./1225., 0., 1./2304.},
	{1., 64./(9.*PI*PI), .25, 64./(225.*PI*PI), 0., 64./(11025.*PI*PI),
		 0., 64./(99225*PI*PI), 0.},
	{1., 81.*PI*PI/1024., 9./25., 9.*PI*PI/1024., 9./1225., 0.,
		 1./11025., 0., 1./148225.}
};

int	RectMaxCoeff = MAXN;
double	*RectCoeff = ToneCoeff[1];


int RectToneInit(void)
{
	int p;
	p = dtoi(Power);
	p = max(min(MAXP, p), 1);
	if (p != Power)
		/* dprintf("Warning. RECT using Power = %d for tones.\n", p)*/;
	RectCoeff = ToneCoeff[p - 1];
/*
	{ int t; 
	for (t = 1; t <= 10; t++)
		printf("|S%d/S0|^2 = %g\n", t,
			RectCoeff[t]);
	}
*/
	return p;
}


/*************************************************************************
	Given a tonal input, perform the CCF after rectification, given that
	*both fibers* are active.
 *************************************************************************/
int ToneELfcm(double SPL, double f0, double ts, double fcm,
			  double *outp, int tup, double tstep)
{
	int n, p, t;
	double coeff[MAXN+1], tau, arg, GSqd();
	Fraction ToneActive();

	p = RectToneInit();			/* Get power of rectifier */
	
	for (n = 1; n <= RectMaxCoeff; n++) // changed const from 1 --> 2
		coeff[n] = 2 * RectCoeff[n] * GSqd(n * f0);

	for (tau = -tup * tstep, t = -tup; t <= tup; t++, tau += tstep) {
		arg = 2.e-3 * PI * f0 * (tau-ts);

		outp[t] = 1;
		
		for (n = 1; n <= MAXN; n++)
			outp[t] += coeff[n] * cos(n * arg);

	/*%%%%%%%%%%%%%%%%%%%%%%%% Manual Rectification %%%%%%%%%%%%%%%%%%%%%%%%*/
		if (outp[t] < 0.0)
			; //outp[t] = 0.0;
		else		// can add ' * pow(10, SPL/10) ' term below				 
			outp[t] *= (DOUBLY_ACTIVE_PROD);
	}
	return p;	/* Return Power of rectifier */
}



/* TOR - function that half-wave rectifies an array of length N. */
void AMRectify(float *x, int N)
{
	int n;
	for (n = 0; n < N; n++) {
		if (x[n] > 0.0)
			x[n] = pow(x[n], Power);
		else
			x[n] = 0.0;
	}
}

